/*
  Copyright 2009 Patrick Ahlbrecht

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/


import java.awt.*;
import javax.xml.parsers.*;
import org.xml.sax.helpers.*;
import org.xml.sax.*;
import java.util.*;

/**
 * Parse an HTML table and transform it into <code>Cell</code>s.
 */
public class TableParser extends DefaultHandler {

  
  /**
   * Collect assembled <code>Cell</code>s;
   */
  private Vector<Cell> cells = new Vector<Cell>();
  
  /**
   * Row counter
   */
  private int row=0;
  
  /**
   * Inner cell padding
   */
  private int cellpadding=0;
  
  /**
   * Outer cell padding
   */
  private Insets cellspacing;
  
  /**
   * The cell, that is currently assembled
   */
  private Cell current = new Cell(null,new GridBagConstraints());
  
  /**
   * HTML constants for cell alignment
   */
  private static final String[][] locations = {
    {"top","left"},
    {"top","center"},
    {"top","right"},
    {"middle","left"},
    {"middle","center"},
    {"middle","right"},
    {"bottom","left"},
    {"bottom","center"},
    {"bottom","right"},
  };
  
  /**
   * GridBag anchors (aligned with <code>location</code>)
   */
  private static final int[] anchors = {
    GridBagConstraints.NORTHWEST,
    GridBagConstraints.NORTH,
    GridBagConstraints.NORTHEAST,
    GridBagConstraints.WEST,
    GridBagConstraints.CENTER,
    GridBagConstraints.EAST,
    GridBagConstraints.SOUTHWEST,
    GridBagConstraints.SOUTH,
    GridBagConstraints.SOUTHEAST
  };
  
  /**
   * Cell alignment
   */
  private String valign = "middle";
  
  /**
   * Cell alignment
   */
  private String halign = "center";
  

  // Constructor
  public TableParser(){}
  
  // Overridden method
  public void startElement(String uri, String localName, String qName, Attributes attributes) {
    if (qName.equalsIgnoreCase("table")) {
      if (attributes!=null) {
        for (int i=0;i<attributes.getLength();i++) {
          String key = attributes.getLocalName(i);
          if (key.equalsIgnoreCase("cellpadding")) {
            cellpadding=Integer.parseInt(attributes.getValue(i));
          }
          if (key.equalsIgnoreCase("cellspacing")) {
            int x=Integer.parseInt(attributes.getValue(i));
            cellspacing = new Insets(x,x,x,x);
          }
        }
      }  
    }
    
    if (qName.equalsIgnoreCase("tr")) {
    }
    
    if (qName.equalsIgnoreCase("td")) {
      
      if (attributes!=null) {
        for (int i=0;i<attributes.getLength();i++) {
          String key = attributes.getLocalName(i);
          if (key.equalsIgnoreCase("rowspan")) {
            current.gbc.gridheight=Integer.parseInt(attributes.getValue(i));
          }
          if (key.equalsIgnoreCase("colspan")) {
            current.gbc.gridwidth=Integer.parseInt(attributes.getValue(i));
          }
          if (key.equalsIgnoreCase("width")) {
            current.gbc.weightx=toFloat(attributes.getValue(i));
          }
          if (key.equalsIgnoreCase("height")) {
            current.gbc.weighty=toFloat(attributes.getValue(i));
          }
          if (key.equalsIgnoreCase("align")) {
            halign=attributes.getValue(i);
          }
          if (key.equalsIgnoreCase("valign")) {
            valign=attributes.getValue(i);
          }
        }
      }
    }
    
  }
  
  // Overridden method
  public void endElement(String uri, String localName, String qName) {
    if (qName.equalsIgnoreCase("table")) {
    }
    
    if (qName.equalsIgnoreCase("tr")) {
      row++;
    }
    
    if (qName.equalsIgnoreCase("td")) {
      current.gbc.anchor = getAnchor();
      if (current.gbc.anchor == GridBagConstraints.CENTER) 
        current.gbc.fill = GridBagConstraints.BOTH;
      else current.gbc.fill = GridBagConstraints.NONE;
      current.gbc.ipadx=cellpadding;
      current.gbc.ipady=cellpadding;
      if (cellspacing!=null) current.gbc.insets=cellspacing;
      current.gbc.gridy=row;
      current.gbc.gridx=nextFree();
      cells.add(current);
      
      
      current = new Cell(null,new GridBagConstraints());
      halign="center";
      valign="middle";
    }

  }
  
  // Overridden method
  public void characters(char[] ch, int start, int length) {
    current.compName=new String(ch,start,length).trim();
  }
  
  /**
   * Determine anchor from alignment
   * @return where to anchor the component in the cell
   */
  private int getAnchor() {
    int ret = GridBagConstraints.CENTER;
    for (int i = 0;i<locations.length;i++) {
      if (locations[i][0].equalsIgnoreCase(valign) && locations[i][1].equalsIgnoreCase(halign)) {
        ret = anchors[i];
      }
    }
    return ret;
  }
  
  /**
   * Find the next free spot in the current row
   * @return the first column in the current row, that is not occupied
   */
  private int nextFree() {
    // This is not the best algorithm to do it!
    Cell[] c = cells.toArray(new Cell[0]);
    int col = 0;
    while (true) {
      if (isFree(col,row,c)) break;
      else col++;
    }
    return col;
  }
  
  /**
   * Check if a cell is already occupied
   * @param x column of the cell
   * @param y row of the cell
   * @param c all cells, that are already placed
   * @return true if x:y is not taken
   */
  private boolean isFree(int x, int y, Cell[] c) {
    for (Cell tmp:c) {
      Rectangle r = new Rectangle(tmp.gbc.gridx, tmp.gbc.gridy, tmp.gbc.gridwidth, tmp.gbc.gridheight);
      if (r.contains(new Point(x,y))) return false;
    }
    return true;
  }
  
  /**
   * Helper function to translate width and height values of a table cell 
   * into floating values
   * @param raw a string like "99%" or "99"
   * @return the float value
   */
  private float toFloat(String raw) {
    String tmp=raw;
    if (raw.endsWith("%")) {
      tmp=raw.substring(0,raw.length()-1);
    }
    return Float.parseFloat(tmp);
  }
  
  /**
   * Get the <code>Cell</code>s, the table consists of
   * @return the properly layed out cells
   */
  public Cell[] getCells() {
    Cell[] tmp = cells.toArray(new Cell[0]);
    return tmp;
  }
}
 